"""Pytest entry point for upto.CurryModel tests."""

from pydantic import BaseModel, Field
from upto import CurryModel


def test_simple_curry_model():
    class MyModel(BaseModel):
        x: str
        y: int
        z: tuple[str, int]

    curried_model1 = CurryModel(MyModel)
    curried_model1(x="1")
    curried_model1(y=2)
    model_instance1 = curried_model1(z=("3", 4))

    curried_model_2 = CurryModel(MyModel)
    model_instance_2 = curried_model_2(x="1")(y=2)(z=("3", 4))

    curried_model_3 = CurryModel(MyModel)
    model_instance_3 = curried_model_3(x="1", y=2)(z=("3", 4))

    assert model_instance1 == model_instance_2 == model_instance_3


def test_curry_default_arg_model_eager():
    class MyModel(BaseModel):
        x: int
        y: int
        z: int = 3

    curried = CurryModel(MyModel)
    x_curried = curried(x=1)
    assert isinstance(x_curried, CurryModel)

    model = x_curried(y=2)
    assert isinstance(model, BaseModel)


def test_curry_default_arg_model_non_eager():
    class MyModel(BaseModel):
        x: int
        y: int
        z: int = 3

    curried = CurryModel(MyModel, eager=False)
    x_curried = curried(x=1)
    assert isinstance(x_curried, CurryModel)

    y_curried = x_curried(y=2)
    assert isinstance(y_curried, CurryModel)

    model = y_curried(z=33)
    assert isinstance(model, BaseModel)
    assert model.z == 33



"""Pytest entry point for init_model_from_kwargs tests."""

import pytest

from tests.data.init_model_from_kwargs_parameters import (
    init_model_from_kwargs_parameters,
)
from upto import init_model_from_kwargs


@pytest.mark.parametrize(("model", "kwargs"), init_model_from_kwargs_parameters)
def test_init_model_from_kwargs(model, kwargs):
    """Check if the init_model_from_kwargs constructor successfully inits a model based on kwargs."""
    for _kwargs in kwargs:
        model_instance = init_model_from_kwargs(model, **_kwargs)
        assert isinstance(model_instance, model)



"""Model and kwargs data mappings for testing upto.init_model_from_kwargs"""

from pydantic import BaseModel
from upto.pydantic_tools.model_constructors import init_model_from_kwargs


class SimpleModel(BaseModel):
    x: int
    y: int = 3


class NestedModel(BaseModel):
    a: str
    b: SimpleModel


class ComplexModel(BaseModel):
    p: str
    q: NestedModel


init_model_from_kwargs_parameters = [
    (SimpleModel, [{"x": 1}, {"x": 1, "y": 2}]),
    (
        NestedModel,
        [
            {"x": 1, "a": "a value"},
            {"x": 1, "y": 2, "a": "a value"},
            {"a": "a value", "b": SimpleModel(x=1)},
            {"a": "a value", "b": SimpleModel(x=1, y=2)},
        ],
    ),
    (
        ComplexModel,
        [
            {"p": "p value", "a": "a value", "x": 1},
            {"p": "p value", "a": "a value", "x": 1, "y": 2},
            {
                "p": "p value",
                "q": init_model_from_kwargs(NestedModel, **{"a": "a value", "x": 1}),
            },
            {"p": "p value", "q": NestedModel(a="a value", b=SimpleModel(x=1))},
        ],
    ),
]



"""Mixin enabling Pydantic models to check for mutually dependent fields."""

from typing import NoReturn, Protocol

from pydantic import BaseModel, ConfigDict, model_validator


class MutualExclusionException(Exception):
    pass


class MutualDependencyException(Exception):
    pass


class _MutualConstraintTemplate(Protocol):
    constraint_id: str

    def condition(self, pair: tuple[str, str], model_data: dict) -> bool: ...

    def throw(self, pair: tuple[str, str]) -> NoReturn: ...


class MutualExclusionTemplate(_MutualConstraintTemplate):
    constraint_id = "mutually_exclusive"

    def condition(self, pair: tuple[str, str], model_data: dict) -> bool:
        return all(field in model_data for field in pair)

    def throw(self, pair: tuple[str, str]) -> NoReturn:
        raise MutualExclusionException(
            f"Fields {' and '.join(pair)} are mutually exclusive."
        )


class MutualDependencyTemplate(_MutualConstraintTemplate):
    constraint_id = "mutually_dependent"

    def condition(self, pair: tuple[str, str], model_data: dict) -> bool:
        return not all(field in model_data for field in pair)

    def throw(self, pair: tuple[str, str]) -> NoReturn:
        raise MutualDependencyException(
            f"Fields {' and '.join(pair)} are mutually dependent."
        )


class _MutualConstraintMixin:
    """Mixin for adding mutual exclusion and dependency validators to a model.

    Subclass _MutualConstrainMixin in a Pydantic model definition
    to add the validator to the model.

    The validator reads two keys from model_config:

    - mutually_exclusive: Iterable[tuple[str, str]]
    - mutually_dependent: Iterable[tuple[str, str]]

    and checks the model fields accordingly.
    """

    constraints = [
        MutualExclusionTemplate(),
        MutualDependencyTemplate(),
    ]

    @model_validator(mode="before")
    @classmethod
    def _check_mutual_constraints(cls, data: dict) -> dict:
        config = cls.model_config  # type: ignore

        for constraint in cls.constraints:
            if pairs := config.get(constraint.constraint_id):
                for pair in pairs:
                    if constraint.condition(pair, data):
                        constraint.throw(pair)

        return data



"""CurryModel: Constructor for currying a Pydantic Model."""

from typing import Any, Self

from pydantic import BaseModel, ValidationError


def validate_model_field(model: type[BaseModel], field: str, value: Any) -> Any:
    """Validate value for a single field given a model.

    Note: Using a TypeVar for value is not possible here,
    because Pydantic might coerce values (if not not in Strict Mode).
    """

    if field not in model.model_fields:
        raise ValueError(f"'{field}' does not denote a field of model '{model}'.")

    try:
        model(**{field: value})
    except ValidationError as e:
        for error in e.errors():
            if field in error["loc"]:
                raise ValidationError.from_exception_data(model.__name__, [error])

    return value


class CurryModel[_TModelInstance: BaseModel]:
    """Constructor for currying a Pydantic Model.

    A CurryModel instance can be called with kwargs which are run against
    the respective model field validators and kept in a kwargs cache.
    Once the model can be instantiated, calling a CurryModel object will
    instantiate the Pydantic model and return the model instance.

    If the eager flag is True (default), model field default values are
    added to the cache automatically, which means that models can be instantiated
    as soon possible, i.e. as soon as all /required/ field values are provided.

    If the fail_fast flag is True (default), the constructor raises a ValueError
    as soon as a value fails to validate against a field constraint.
    """

    def __init__(
        self, model: type[_TModelInstance], eager: bool = True, fail_fast: bool = True
    ) -> None:
        self.model = model
        self.eager = eager
        self.fail_fast = fail_fast

        self._kwargs_cache: dict = (
            {k: v.default for k, v in model.model_fields.items() if not v.is_required()}
            if eager
            else {}
        )

    def __repr__(self):  # pragma: no cover
        return f"CurryModel object {self._kwargs_cache}"

    def __call__(self, **kwargs: Any) -> Self | _TModelInstance:
        if self.fail_fast:
            for k, v in kwargs.items():
                validate_model_field(self.model, k, v)

        self._kwargs_cache.update(kwargs)

        if self.model.model_fields.keys() == self._kwargs_cache.keys():
            return self.model(**self._kwargs_cache)
        return self



"""Basic test for upto.validate_model_field."""

import pytest

from pydantic import BaseModel
from upto import validate_model_field


def test_validate_model_field_sad_path_missing_field():
    class Model(BaseModel):
        x: int

    with pytest.raises(ValueError):
        validate_model_field(model=Model, field="dne", value=None)


def test_validate_model_field_sad_path_invalid_value():
    class Model(BaseModel):
        x: int

    with pytest.raises(ValueError):
        validate_model_field(model=Model, field="x", value=1.1)



"""Basic tests for upto.ichunk."""

from collections.abc import Iterator
from typing import NamedTuple

import pytest
from upto import ichunk


class IChunkTestParameter(NamedTuple):
    iterator: Iterator
    size: int
    expected: list[tuple]


params = [
    IChunkTestParameter(
        iterator=iter(range(5)), size=2, expected=[(0, 1), (2, 3), (4,)]
    ),
    IChunkTestParameter(
        iterator=iter(range(6)), size=2, expected=[(0, 1), (2, 3), (4, 5)]
    ),
    IChunkTestParameter(
        iterator=iter(range(6)),
        size=3,
        expected=[(0, 1, 2), (3, 4, 5)],
    ),
    IChunkTestParameter(iterator=iter(range(0)), size=2, expected=[]),
    IChunkTestParameter(iterator=iter(range(5)), size=5, expected=[(0, 1, 2, 3, 4)]),
    IChunkTestParameter(iterator=iter(range(5)), size=50, expected=[(0, 1, 2, 3, 4)]),
    IChunkTestParameter(iterator=iter(range(5)), size=0, expected=[]),
]


@pytest.mark.parametrize("param", params)
def test_ichunk(param):
    chunks = ichunk(iterator=param.iterator, size=param.size)
    result = [tuple(chunk) for chunk in chunks]

    assert result == param.expected



"""ComposeRouter: Utility for routing methods through a functional pipeline."""

from collections.abc import Callable

from toolz import compose


class ComposeRouter:
    """ComposeRouter.

    This class routes attributes access for registered methods
    through a functional pipeline constructed from components.

    Example:

    class Foo:
        route = ComposeRouter(lambda x: x + 1, lambda y: y * 2)

        @route.register
        def method(self, x, y):
            return x * y

    foo = Foo()

    print(foo.method(2, 3))           # 6
    print(foo.route.method(2, 3))     # 13
    """

    def __init__(self, *components: Callable) -> None:
        self._components: tuple[Callable, ...] = components
        self._registry: list = []

    def register[F: Callable](self, f: F) -> F:
        self._registry.append(f.__name__)
        return f

    def __get__[T](self, instance: T, owner: type[T]):
        class _wrapper:
            def __init__(_self, other):
                _self.other = other

            def __getattr__(_self, name):
                if name in self._registry:
                    method = getattr(_self.other, name)
                    return compose(*self._components, method)

                raise AttributeError(f"Name '{name}' not registered.")

        return _wrapper(instance)



"""Pytest entry point for upto.ComposeRouter tests."""

import pytest

from toolz import compose
from upto import ComposeRouter


def test_simple_compose_router():
    """Simple base test case for upto.ComposeRouter.

    Check if the composed route generates the same result
    as the _components applied to the result of the method without routing.
    """
    _components = [lambda x: x + 1, lambda y: y * 2]

    class Foo:
        route = ComposeRouter(*_components)

        @route.register
        def method(self, x, y):
            return x * y

    foo = Foo()

    no_route = foo.method(2, 3)
    assert foo.route.method(2, 3) == compose(*_components)(no_route)


def test_simple_compse_router_unregistered_fail():
    _components = [lambda x: x + 1, lambda y: y * 2]

    class Foo:
        route = ComposeRouter(*_components)

        def method(self, x, y):
            return x * y

    foo = Foo()

    with pytest.raises(AttributeError):
        foo.route.method(2, 3)



"""Pytest entry point for mutual constraint validation mixin."""

from typing import NamedTuple

import pytest

from pydantic import BaseModel, ConfigDict
from upto import _MutualConstraintMixin
from upto.pydantic_tools.mutual_constraint_validator import (
    MutualDependencyException,
    MutualExclusionException,
)


class Model(BaseModel, _MutualConstraintMixin):
    a: int | None = None
    b: int | None = None
    c: int | None = None
    d: int | None = None


class MutualConstraintParameter(NamedTuple):
    model: BaseModel
    model_config: ConfigDict
    model_init: dict
    exception: Exception | None = None


mutual_constraint_pass_parameters = [
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(),
        model_init={"a": 1, "c": 3},
        exception=None,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_exclusive=[("a", "b")]),
        model_init={"a": 1, "c": 3},
        exception=None,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_exclusive=[("a", "b")]),
        model_init={"b": 2, "c": 3, "d": 4},
        exception=None,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_exclusive=[("a", "d")]),
        model_init={"a": 1, "b": 2, "c": 3},
        exception=None,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_dependent=[("a", "b")]),
        model_init={"a": 1, "b": 2, "c": 3},
        exception=MutualDependencyException,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_dependent=[("a", "b")]),
        model_init={"c": 3, "d": 4, "a": 1, "b": 2},
        exception=MutualDependencyException,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_dependent=[("a", "d")]),
        model_init={"a": 1, "c": 3, "d": 4},
        exception=MutualDependencyException,
    ),
]

mutual_constraint_fail_parameters: list[MutualConstraintParameter] = [
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_exclusive=[("a", "b")]),
        model_init={"a": 1, "b": 2},
        exception=MutualExclusionException,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_exclusive=[("a", "b")]),
        model_init={"a": 1, "b": 2, "c": 3, "d": 4},
        exception=MutualExclusionException,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_exclusive=[("a", "d")]),
        model_init={"a": 1, "b": 2, "c": 3, "d": 4},
        exception=MutualExclusionException,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_dependent=[("a", "b")]),
        model_init={"a": 1, "c": 3},
        exception=MutualDependencyException,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_dependent=[("a", "b")]),
        model_init={"c": 3, "d": 4},
        exception=MutualDependencyException,
    ),
    MutualConstraintParameter(
        model=Model,
        model_config=ConfigDict(mutually_dependent=[("a", "d")]),
        model_init={"c": 3, "d": 4},
        exception=MutualDependencyException,
    ),
]


@pytest.mark.parametrize(
    MutualConstraintParameter._fields, mutual_constraint_pass_parameters
)
def test_happy_path_mutual_constrain_validator(
    model, model_config, model_init, exception
):
    model.model_config = model_config
    assert model(**model_init)


@pytest.mark.parametrize(
    MutualConstraintParameter._fields, mutual_constraint_fail_parameters
)
def test_sad_path_mutual_constrain_validator(
    model, model_config, model_init, exception
):
    model.model_config = model_config

    with pytest.raises(exception):
        model(**model_init)



"""Pydantic model constructors."""

from pydantic import BaseModel


def init_model_from_kwargs[ModelType: BaseModel](
    model: type[ModelType], **kwargs
) -> ModelType:
    """Instantiate a (potentially nested) model from (flat) kwargs.

    Example:

        class SimpleModel(BaseModel):
            x: int
            y: int


        class NestedModel(BaseModel):
            a: str
            b: SimpleModel


        class ComplexModel(BaseModel):
            p: str
            q: NestedModel


    model = instantiate_model_from_kwargs(ComplexModel, x=1, y=2, a="a value", p="p value")
    print(model)  # p='p value' q=NestedModel(a='a value', b=SimpleModel(x=1, y=2))
    """

    def _model_class_p(field_key, field_value, kwargs) -> bool:
        """Helper for _get_bindings.

        Check if a field is annoted with a model type and there is no applicable model instance kwarg.
        This triggers the recursive case in _get_bindings.
        """
        return isinstance(field_value.annotation, type(BaseModel)) and not isinstance(
            kwargs.get(field_key), BaseModel
        )

    def _construct_bindings(model: type[ModelType], **kwargs) -> dict:
        """Construct bindings needed for model instantiation from kwargs.

        If a field is annotated with a model type and there is no applicable model instance kwarg,
        trigger the recursive clause and consult kwargs for the nested model.
        """
        return {
            k: (
                v.annotation(**_construct_bindings(v.annotation, **kwargs))
                if _model_class_p(k, v, kwargs)
                else kwargs.get(k, v.default)
            )
            for k, v in model.model_fields.items()
        }

    return model(**_construct_bindings(model, **kwargs))



"""A simple chunk iterator."""

from collections.abc import Iterator
from itertools import chain, islice
from typing import cast


def ichunk[T](iterator: Iterator[T], size: int) -> Iterator[Iterator[T]]:
    _missing = object()
    chunk = islice(iterator, size)

    if (first := next(chunk, _missing)) is _missing:
        return

    yield chain[T]([cast(T, first)], chunk)
    yield from ichunk(iterator, size=size)
